#include <linux/kernel.h>
#include <linux/string.h>

#include <asm/console.h>
#include <asm/hwrpb.h>
#include <asm/system.h>

#include "aboot.h"
#include "cons.h"

long cons_dev;			/* console device */


long cons_puts(const char *str, long len)
{
	long remaining, written;

	for (remaining = len; remaining; remaining -= written) {
		written = dispatch(CCB_PUTS, cons_dev, str, remaining);
		if (written < 0) {
			halt();		/* better than random crash... */
		}
		str += written;
	}
	return len;
}

void cons_putchar(char c)
{
  char buf[2];

  buf[0] = c;
  buf[1] = 0;
  cons_puts(buf,1);
}

int cons_getchar(void)
{
	long c;

  while ((c = dispatch(CCB_GETC, cons_dev)) < 0)
  	;
  return c;
}


long cons_getenv(long index, char *envval, long maxlen)
{
	long len;

	len = dispatch(CCB_GET_ENV, index, envval, maxlen - 1);
	if (len >= 0) {
		envval[len] = '\0';
	}
	return len;
}


long cons_open(const char *devname)
{
	return dispatch(CCB_OPEN, devname, strlen(devname));
}


long cons_close(long dev)
{
	return dispatch(CCB_CLOSE, dev);
}


long cons_read(long dev, void *buf, long count, long offset)
{
	static char readbuf[SECT_SIZE];		/* minize frame size */

	if ((count & (SECT_SIZE-1)) == 0 && (offset & (SECT_SIZE-1)) == 0) {
		/* I/O is aligned... this is easy! */
		return dispatch(CCB_READ, dev, count, buf,
				offset / SECT_SIZE);
	} else {
		long bytesleft, iocount, blockoffset, iosize, lbn, retval;

		bytesleft = count;
		iocount = 0;
		blockoffset = offset % SECT_SIZE;
		lbn = offset / SECT_SIZE;

		while (bytesleft > 0) {
			if ((blockoffset == 0) && (bytesleft >= SECT_SIZE)) {
				/*
				 * This portion of the I/O is aligned,
				 * so read it straight in:
				 */
				iosize = SECT_SIZE;
				retval = dispatch(CCB_READ, dev, iosize, buf,
						  lbn);
				if (retval != iosize) {
					printk("read error 0x%lx\n",retval);
					return -1;
				}
			} else {
				/*
				 * Not aligned; must read it into a
				 * temporary buffer and go from there.
				 */
				retval = dispatch(CCB_READ, dev, SECT_SIZE,
						  readbuf, lbn);
				if (retval != SECT_SIZE) {
					printk("read error, lbn %ld: 0x%lx\n",
					       lbn, retval);
					return -1;
				}
				iosize = bytesleft;
				if (blockoffset + iosize >= SECT_SIZE) {
					iosize = SECT_SIZE - blockoffset;
				}
				memcpy(buf, readbuf + blockoffset, iosize);
			}
			buf += iosize;
			iocount += iosize;
			bytesleft -= iosize;
			blockoffset = 0;
			++lbn;
		}
		return iocount;
	}
}


void cons_init(void)
{
	char envval[256];

	if (cons_getenv(ENV_TTY_DEV, envval, sizeof(envval)) < 0) {
		halt();		/* better than random crash */
	}
	cons_dev = simple_strtoul(envval, 0, 10);
}
