/*
 * aboot/main.c
 *
 * Copyright (C) 1994, 1995, 1996 Linus Torvalds
 *
 * This file is the bootloader for the Linux/AXP kernel.
 *
 * Filesystem support added by David Mosberger (davidm@cs.arizona.edu).
 *
 * commandline argument handling by Michael Schwingen 
 * (rincewind@discworld.oche.de)
 */
#include <linux/kernel.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/a.out.h>

#include <asm/console.h>
#include <asm/hwrpb.h>
#include <asm/system.h>

#include <alloca.h>

#include "aboot.h"
#include "bootfs.h"
#include "cons.h"
#include "disklabel.h"
#include "setjmp.h"
#include "utils.h"

extern struct bootfs ext2fs;
extern struct bootfs ufs;
extern struct bootfs dummyfs;

static struct bootfs *bootfs[] = {
	&ext2fs,
	&ufs
};

struct bootfs *bfs = 0;		/* filesystem to boot from */
char *dest_addr = 0;
long bytes_to_copy = 0;
jmp_buf jump_buffer;

char boot_file[256], kernel_args[256] = "";
struct disklabel *label = 0;

static char *bss_start = 0;
static long bss_size = 0;			/* size of bss */
static unsigned long start_addr = START_ADDR;
static unsigned long entry_addr = START_ADDR;
static struct disklabel dummy_label = {
	DISKLABELMAGIC,		/* d_magic */
	0, 0,			/* d_type, d_subtype */
	"dummy",		/* d_typename[16] */
	"none",			/* d_packname[16] */
	512,			/* d_secsize */
	2880,			/* d_nsectors */
	80,			/* d_ntracks */
	80,			/* d_ncylinders */
	36,			/* d_secpercyl */
	0,			/* d_secprtunit */
	0,			/* d_sparespertrack */
	0,			/* d_sparespercyl */
	0,			/* d_acylinders */
	0, 0, 0, 0,		/* d_rpm, d_interleave,
				   d_trackskew, d_cylskew */
	0, 0, 0,		/* d_headswitch, d_trkseek,
				   d_flags */
	{ 0 },			/* d_drivedata[5] */
	{ 0 },			/* d_spare[5] */
	DISKLABELMAGIC,		/* d_magic2 */
	0,			/* d_checksum */
	1,			/* d_npartitions */
	0, 0,			/* d_bbsize, d_sbsize */
	{{
		~0U,		/* p_size */
		0,		/* p_offset: starting sector */
		512,		/* p_fsize: fragment size */
		FS_EXT2,	/* p_fstype: filessytem type */
		0,		/* p_frag: fragments/block */
		0		/* p_cpg: cyclinders/group */
	}}
};


#ifdef DEBUG
 void wait_key(void);
#endif


long
first_block (const char *buf, long blocksize)
{
	struct exec *ex;
	long text_offset, nbytes;

	ex = (struct exec*) buf;
	if (N_MAGIC(*ex) != OMAGIC) {
		printk("aboot: unexpected object file format %lo\n",
		       N_MAGIC(*ex));
		return -1;
	}

	start_addr = N_TXTADDR(*ex);
	if (start_addr < 0xfffffc0000000000UL) {
		printk("aboot: bad kernel start address %lx\n",
		       N_TXTADDR(*ex));
		return -1;
	}

	entry_addr    = ex->a_entry;
	text_offset   = N_TXTOFF(*ex);
	dest_addr     = (char *) start_addr;
	bytes_to_copy = ex->a_text + ex->a_data;
	bss_start     = dest_addr + bytes_to_copy;
	bss_size      = ex->a_bss;

	nbytes = blocksize - text_offset;
	if (nbytes > 0) {
		memcpy(dest_addr, buf + text_offset, nbytes);
		dest_addr += nbytes;
		bytes_to_copy -= nbytes;
	} else if (nbytes < 0) {
		printk("aboot: block-size smaller than anticipated!\n");
		return -1;
	}
	return 0;
}


/*
 * Attempt a raw boot (header-less kernel follows right after aboot).
 */
int
load_raw (long dev)
{
	printk("aboot: loading raw kernel...\n");

	return dispatch(CCB_READ, dev, CONFIG_KERNEL_SIZE, start_addr,
			CONFIG_ABOOT_SIZE / SECT_SIZE + BOOT_SECTOR);
}


int
load_aout (int fd)
{
	long nread, nblocks;
	unsigned char *buf;

	buf = malloc(bfs->blocksize);

	/* read a.out header: */

	nread = (*bfs->bread)(fd, 0, 1, buf);
	if (nread != bfs->blocksize) {
		printk("aboot: read returned %ld instead of %ld bytes\n",
		       nread, sizeof(buf));
		return -1;
	}
#ifdef DEBUG
	if (1) {
		int i,j,c;

		for(i = 0; i < 16; i++) {
			for (j = 0; j < 16; j++)
				printk("%02X ", buf[j+16*i]);
			for(j = 0; j < 16; j++) {
				c = buf[j+16*i];
				printk("%c", (c >= ' ') ? c : ' ');
			}
			printk("\n");
		}
		wait_key();
	}
#endif
	if (first_block(buf, bfs->blocksize) < 0) {
		return -1;
	}

	/* read rest of kernel in one big swoop: */
	nblocks = (bytes_to_copy + bfs->blocksize - 1) / bfs->blocksize;
	nread = (*bfs->bread)(fd, 1, nblocks, dest_addr);
	if (nread != nblocks * bfs->blocksize) {
		printk("aboot: read returned %ld instead of %ld bytes\n",
		       nread, nblocks * bfs->blocksize);
		return -1;
	}
	return 0;
}


static long
read_kernel (const char *filename)
{
	volatile int attempt, method;
	long len;
	int fd;
	static struct {
		const char *name;
		int (*func)(int fd);
	} read_method[]= {
		{"plain",	load_aout},
		{"compressed",	uncompress_kernel}
	};
	long res;
#	define NUM_METHODS	(sizeof(read_method)/sizeof(read_method[0]))

#ifdef DEBUG
	printk("read_kernel(%s)\n", filename); wait_key();
#endif

	method = 0;
	len = strlen(filename);
	if (len > 3 && filename[len - 3] == '.'
	    && filename[len - 2] == 'g' && filename[len - 1] == 'z')
	{
		/* if filename ends in .gz we don't try plain method: */
		method = 1;
	}

	for (attempt = 0; attempt < NUM_METHODS; ++attempt) {
		fd = (*bfs->open)(filename);
		if (fd < 0) {
			printk("%s: file not found\n", filename);
			return -1;
		}
		printk("aboot: loading %s %s...\n",
		       read_method[method].name, filename);

		if (!_setjmp(jump_buffer)) {
			res = (*read_method[method].func)(fd);
			(*bfs->close)(fd);
			if (res >= 0) {
				return 0;
			}
		}
		method = (method + 1) % NUM_METHODS;
	}
	return -1;
}


static void
get_disklabel (long dev)
{
	static char lsect[512];
	long nread;

#ifdef DEBUG
	printk("load_label(dev=%lx)\n", dev); wait_key();
#endif
	nread = cons_read(dev, &lsect, LABELOFFSET + sizeof(*label),
			  LABELSECTOR);
	if (nread != LABELOFFSET + sizeof(*label)) {
		printk("aboot: read of disklabel sector failed (nread=%ld)\n",
		       nread);
		return;
	}
	label = (struct disklabel*) &lsect[LABELOFFSET];
	if (label->d_magic  == DISKLABELMAGIC &&
	    label->d_magic2 == DISKLABELMAGIC)
	{
		printk("aboot: valid disklabel found: %d partitions.\n",
		       label->d_npartitions);
	} else {
		/*
		 * No disklabel is present.  Create a dummy label
		 * that will allow us to mount an ext2 filesystem
		 * provided it starts at the first sector.  This
		 * case occurs with floppy disks, for example.
		 */
		printk("aboot: no disklabel found.\n");
		label = &dummy_label;
	}
}


struct bootfs *
mount_fs (long dev, int partition)
{
	struct d_partition *part;
	struct bootfs *fs = 0;
	int i, secsize;
	u32 p_offset = 0;

#ifdef DEBUG
	printk("mount_fs(%lx, %d)\n", dev, partition); wait_key();
#endif
	secsize = label->d_secsize;
	if (partition == 0) {
	    fs = &dummyfs;
	} else {
	    if (partition <= 0 || partition > label->d_npartitions) {
		    printk("aboot: invalid partition %u\n", partition);
		    return 0;
	    }
	    part = &label->d_partitions[partition - 1];
	    for (i = 0; bootfs[i]->fs_type != part->p_fstype; ++i) {
		    if (i + 1 >= sizeof(bootfs)/sizeof(bootfs[0])) {
			    printk("aboot: don't know how to mount "
				   "partition %d (filesystem type %d)\n",
				   partition, part->p_fstype);
			    return 0;
		    }
	    }
	    fs = bootfs[i];
	    p_offset = part->p_offset;
	}
	if ((*fs->mount)(dev, p_offset * secsize) < 0) {
		printk("aboot: error when mounting partition %d\n", partition);
		return 0;
	}
	return fs;
}

#ifdef DEBUG

void wait_key (void)
{
	while (cons_getchar() != 13) /* skip */;
}

void
test_key (void)
{
	int c;
	do {
		c = cons_getchar();
		if (c != 0) printk("%d ",c);
	} while (c != 13);
}

#endif /* DEBUG */


void
getline (char *buf, int maxlen)
{
	int len=0;
	char c;

	do {
		c = cons_getchar();
		switch (c) {
		      case 0:
		      case 10:
		      case 13:
			break;
		      case 8:
		      case 127:
			if (len > 0) {
				--len;
				cons_putchar(8);
				cons_putchar(' ');
				cons_putchar(8);
			}
			break;

		      default:
			if (len < maxlen-1 && c >= ' ') {
				buf[len] = c;
				len++;
				cons_putchar(c);
			}
			break;
		}
	} while (c != 13 && c != 10);
	buf[len] = 0;
}


void
print_config_file (struct bootfs *fs)
{
	int fd, nread, blkno = 0;
	char *buf;

	fd = (*fs->open)(CONFIG_FILE);
	if (fd < 0) {
		printk("%s: file not found\n", CONFIG_FILE);
		return;
	}
	buf = malloc(fs->blocksize + 1);
	if (!buf) {
		printk("aboot: malloc failed!\n");
		return;
	}
	do {
		nread = (*fs->bread)(fd, blkno++, 1, buf);
		buf[nread] = '\0';
		printk("%s", buf);
	} while (nread > 0);
	(*fs->close)(fd);
}


int
get_default_args (struct bootfs *fs, char *str, int num)
{
	int fd, nread, state, line, blkno = 0;
	char *buf, *d, *p;

#ifdef DEBUG
	printk("get_default_args(%s,%d)\n", str, num); wait_key();
#endif
	*str = '\0';
	fd = (*fs->open)(CONFIG_FILE);
	if (fd < 0) {
		printk("%s: file not found\n", CONFIG_FILE);
		return -1;
	}
	buf = malloc(fs->blocksize);
	if (!buf) {
		printk("aboot: malloc failed!\n");
		return -1;
	}
	d = str;
	line = 1;
	state = 2;
	do {
		nread = (*fs->bread)(fd, blkno++, 1, buf);
		p = buf;
		while (p < buf + nread && *p && state != 5) {
			switch (state) {
			      case 0: /* ignore rest of line */
			      case 1: /* in comment */
				if (*p == '\n') state = 2;
				break;

			      case 2: /* after end of line */
				line++;
				if (*p == num) {
					state = 3;	/* found it... */
					break;
				}
				if (*p == '#') {
					state = 1;	/* comment */
					break;
				}
				if (*p == '-') {
					state = 5;	/* end-of-file mark */
					break;
				}
				state = 0;	/* ignore rest of line */
				break;

			      case 3: /* after matched number */
				if (*p == ':') {
					state = 4;	/* copy string */
				} else {
					state = 2;	/* ignore rest */
					printk("aboot: syntax error in line "
					       "%d: `:' expected\n", line);
				}
				break;

			      case 4: /* copy until EOL */
				if (*p == '\n') {
					*d = 0;
					state=5;
				} else {
					*d++ = *p;
				}
				break;

			      default:
			}
			p++;
		}
	} while (nread > 0 && state != 5);
	(*fs->close)(fd);
#ifdef DEBUG
	printk("get_default done\n"); wait_key();
#endif

	if (state != 5) {
		printk("aboot: could not find default config `%c'\n", num);
		return -1;
	}
	return 0;
}


void
get_options (long dev)
{
	long result;
	char buf[256],*p;
	int opt;

#ifdef DEBUG
	printk("get_options(%lx)\n",dev); wait_key();
#endif
	result = dispatch(CCB_GET_ENV, ENV_BOOTED_FILE, boot_file, 255);
	if (result < 0) {
		printk("aboot: warning: can't get ENV_BOOTED_FILE!\n");
		result = 0;
	}
	boot_file[result] = '\0';
	result = dispatch(CCB_GET_ENV, ENV_BOOTED_OSFLAGS, kernel_args, 255);
	if  (result < 0)
		result = 0;	/* failed to read BOOTED_OSFLAGS */
	kernel_args[result] = '\0';

	if (kernel_args[0] && kernel_args[1] == '\0') {
		int part = (label == &dummy_label) ? 1 : CONFIG_FILE_PARTITION;
		struct bootfs *fs;

		opt = kernel_args[0];
		kernel_args[0] = '\0';

		fs = mount_fs(dev, part);
		if (!fs)
			return;

		do {
			if (opt >= '0' && opt <= '9') {
				if (get_default_args(fs, buf, opt) >= 0)
					break;
				opt = 'i';	/* failed---ask user */
			} else if (opt == 'i' || opt == 'h') {
				if (opt == 'h') {
					print_config_file(fs);
				}
				printk("aboot> ");
				getline(buf, 256);
				printk("\n");
				opt = 0;
				if (buf[0] && buf[1] == '\0')
					opt = buf[0];
				else if (buf[0] == '\0')
					opt = 'i';
			}
		} while ((opt >= '0' && opt <= '9') ||
			 opt == 'i' || opt == 'h');

		p = strchr(buf, ' ');
		if (p) {
			strcpy(kernel_args, p + 1);
			*p = '\0';
		}
		strcpy(boot_file, buf);
#ifdef DEBUG
		printk("\nfile=`%s',args=`%s'\n", boot_file, kernel_args);
#endif
	}
	printk("aboot: options: `%s'\n", kernel_args);
}


static long
load (long dev)
{
	char *fname;
	int part = 1;

#ifdef DEBUG
	printk("load(%lx)\n", dev); wait_key();
#endif
	fname = boot_file;
	if (fname[0] == '\0' || (fname[0] == '-' && fname[1] == '\0')) {
		/* empty filename or a single "-" implies raw boot: */
		return load_raw(dev);
	}
	if (fname[0] >= '0' && fname[0] <= '9' && fname[1] == '/') {
		part = fname[0] - '0';
		fname += 2;
	}
	bfs = mount_fs(dev, part);
	if (!bfs) {
		printk("aboot: mount of partition %d failed, "
		       "attempting raw boot\n", part);
		return load_raw(dev);
	}
	if (read_kernel(fname) < 0) {
		return -1;
	}
	/* clear bss: */
	memset((char*)bss_start, 0, bss_size);
	return 0;
}


/*
 * Start the kernel.
 */
static void
runkernel(void)
{
	__asm__ __volatile__(
		"bis %1,%1,$30\n\t"
		"bis %0,%0,$26\n\t"
		"ret ($26)"
		: /* no outputs: it doesn't even return */
		: "r" (entry_addr),
		  "r" (PAGE_SIZE + INIT_STACK));
}


/* this is started by the startup code */
void
start_kernel(void)
{
	long i, dev;
	char envval[256];

	cons_init();

	printk("aboot: Linux/Alpha bootloader version 0.3\n");

	if (INIT_HWRPB->pagesize != 8192) {
		printk("aboot: expected 8kB pages, got %ldkB\n",
		       INIT_HWRPB->pagesize >> 10);
		return;
	}

	pal_init();
	if (cons_getenv(ENV_BOOTED_DEV, envval, sizeof(envval)) < 0) {
		printk("aboot: Can't get BOOTED_DEV environment variable!\n");
		return;
	}
	dev = cons_open(envval);
	if (dev < 0) {
		printk("aboot: unable to open boot device `%s': %lx\n",
		       envval, dev);
		return;
	}
	dev &= 0xffffffff;
	get_disklabel(dev);
	get_options(dev);
	i = load(dev);
#ifdef DEBUG
	printk("load done\n"); wait_key();
#endif	
	cons_close(dev);
	if (i < 0) {
		printk("aboot: load failed (%lx)\n", i);
		return;
	}
	strcpy((char*)ZERO_PGE, kernel_args);
	printk("aboot: ok, now starting the kernel...\n");
	runkernel();
	printk("aboot: kernel returned unexpectedly.  Halting slowly...\n");
	for (i = 0 ; i < 0x100000000 ; i++)
		/* nothing */;
	halt();
}
